
import asyncio
import collections
import base64
import datetime
import numpy as np
import PIL.Image
import io
import time


def _make_preview(image, vmin, vmax, preview_size = (128, 128)):
    """Convert an image to an 8-bit RGB image"""
    channels = 1 if image.ndim == 2 else image.shape[2]
    
    if channels != 1 and channels != 3:
        image = image[0]
    
    if channels == 1 and image.ndim == 3:
        image = np.squeeze(image)

    im = 255 * (image.astype(np.float) - vmin) / (vmax - vmin)

    im = PIL.Image.fromarray(im.astype(np.uint8))
    im = im.resize(preview_size, PIL.Image.BILINEAR)
    data = io.BytesIO()
    im.save(data, format = "jpeg")
    data.seek(0)
    return data.getvalue()


ImageValues = collections.namedtuple("ImageRange", ["vmin", "vmax"])


class Image(object):
    def __init__(self, name, image, values):
        self.id       = 0
        self.name     = name
        self.width    = image.shape[1]
        self.height   = image.shape[0]
        self.channels = 1 if image.ndim == 2 else image.shape[2]
        self.data     = image.tobytes()
        self.dtype    = image.dtype.name
        self.values   = values
        self.time     = datetime.datetime.now()
                
        self.preview = _make_preview(image, values.vmin, values.vmax)

    def info(self):
        return dict(id       = self.id,
                    name     = self.name,
                    width    = self.width,
                    height   = self.height,
                    channels = self.channels,
                    dataType = self.dtype,
                    values   = dict(vmin = self.values.vmin, 
                                    vmax = self.values.vmax),
                    time     = str(self.time))

    @staticmethod
    def from_data(data):
        d = base64.b64decode(data["image"])
        image = np.frombuffer(d, dtype = data["dataType"]).reshape(data["height"], data["width"], data["channels"])
        return Image(data["name"], image, ImageValues(vmin = data["values"]["vmin"], vmax = data["values"]["vmax"]))


    @staticmethod
    def create_data(name, image, values):
        return dict(name     = name,
                    width    = image.shape[1],
                    height   = image.shape[0],
                    channels = 1 if image.ndim == 2 else image.shape[2],
                    dataType = image.dtype.name,
                    values   = dict(vmin = values.vmin, 
                                    vmax = values.vmax),
                    image    = base64.b64encode(image.tobytes()).decode("utf-8"))




class Messages(object):
    @staticmethod
    def add_image(image):
        return dict(type = "add_image",
                    info = image.info())

    @staticmethod
    def remove_image(name):
        return dict(type = "remove_image",
                    name = name)

    @staticmethod
    def remove_all_images():
        return dict(type = "remove_all_images")



class ImageDB(object):
    def __init__(self):
        self._id = 0
        self.images = {}
        self.observer = []
        self.wait_time = 0

        self.notify_wait_time()

    def add_image(self, image):
        image.id = self._next_id()
        self.images[image.name] = image
        self.notify_observer(Messages.add_image(image))
        return image.id
        
    def remove_image(self, name):
        if name in self.images:
            self.images.pop(name)
        self.notify_observer(Messages.remove_image(name))

    def remove_all_images(self):
        self.images.clear()
        self.notify_observer(Messages.remove_all_images())


    def get_image(self, name):
        im = self.images.get(name)
        return im

    def _next_id(self):
        self._id += 1
        return self._id

    def notify_wait_time(self):
        self.wait_time = time.time()

    def is_wait_time_ready(self, time):
        return self.wait_time > time

    def add_observer(self, f):
        self.observer.append(f)
        for im in self.images.values():
            f(Messages.add_image(im))

    def remove_observer(self, f):
        self.observer.remove(f)

    def notify_observer(self, msg):
        for f in self.observer:
            f(msg)

    def message_stream(self):
        return MessageStream(self)




class MessageStream(object):
    def __init__(self, db):
        self._db = db
        self._messages = asyncio.Queue()

    def __enter__(self):
        self._db.add_observer(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._db.remove_observer(self)

    def __call__(self, msg):
        self._messages.put_nowait(msg)

    def __aiter__(self):
        return self

    async def __anext__(self):
        return await self._messages.get()








import aiohttp.web
import asyncio
import json
import os.path
import threading
import multiprocessing
import argparse
import time


from . import config
from .service import Image, ImageValues, ImageDB, Messages, MessageStream

ImageNotFound = aiohttp.web.HTTPNotFound()

_db = None

routes = aiohttp.web.RouteTableDef()

def get_image(name):
    im = _db.get_image(name)
    if not im:
        raise ImageNotFound
    return im


@routes.get("/api/preview/{name}")
def preview(request):
    name = request.match_info["name"]
    im = get_image(name)
    return aiohttp.web.Response(body = im.preview, content_type = "image/jpeg")


@routes.get("/api/image/{name}")
def image(request):
    name = request.match_info["name"]
    im = get_image(name)
    return aiohttp.web.Response(body = im.data, content_type = "application/octet-stream")


@routes.post('/api/image')
async def image(request):
    d = await request.content.read()
    data = json.loads(d)
    id = _db.add_image(Image.from_data(data))
    return aiohttp.web.json_response(dict(id = id))


@routes.delete('/api/image/{name}')
def images(request):
    name = request.match_info["name"]
    _db.remove_image(name)
    return aiohttp.web.Response()


@routes.get("/api/images")
def images(request):
    info = [im.info() for im in _db.images.values()]
    return aiohttp.web.json_response({"images": info})


@routes.delete('/api/images')
def images(request):
    _db.remove_all_images()
    return aiohttp.web.Response()

   
@routes.get("/api/image_stream")
async def image_stream(request):
    ws = aiohttp.web.WebSocketResponse()
    await ws.prepare(request)
    
    with _db.message_stream() as stream:
        async for msg in stream:
            await ws.send_json(msg)

    if not ws.closed():
        await ws.close()

    return ws


@routes.get("/api/wait/{time}")
async def wait(request):
    wait_time = int(request.match_info["time"])
    if wait_time == 0:
        wait_time = int(time.time())
    for i in range(10):
        if _db.is_wait_time_ready(wait_time):
            return aiohttp.web.json_response({"ready": True})
        else:
            wait_time = int(time.time())

        await asyncio.sleep(1)
        
    return aiohttp.web.json_response({"ready": False, "time": wait_time})


@routes.post('/api/wait')
def image(request):
    _db.notify_wait_time()
    return aiohttp.web.Response()


@routes.get("/{path:.*}")
async def index(request):
    # This route makes sure that unknown routes are mapped to index.html, so Angular can grab them"""
    path = config.wwwroot() + request.path
    if os.path.isfile(path):
        return  aiohttp.web.FileResponse(path)
    else:
        return aiohttp.web.FileResponse(config.wwwroot() + "/index.html")
    
routes.static("/", config.wwwroot())



def run_server(port = None):
    """Runs the server (blocking)"""

    global _db
    _db = ImageDB()

    if not port:
        port = config.port
    else:
        config.port = port

    app = aiohttp.web.Application()
    app.add_routes(routes)
    aiohttp.web.run_app(app, port = port, shutdown_timeout = 1, print = None)



def run_server_thread(port = None, daemon = True):
    """Runs the server in a new process"""
    def run():
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        run_server(port)

    t = threading.Thread(target = run, daemon = daemon)
    t.start()
    return t


def run_server_process(port = None, daemon = True):
    """Runs the server in a multiprocessing process"""
    if port:
        config.port = port

    p = multiprocessing.Process(target = run_server, args = (port,), daemon = daemon)
    p.start()
    return p





import urllib
import json
import numpy as np

from . import config
from .service import Image, ImageValues

def do_request(url, method, data = None):
    try:
        return urllib.request.urlopen(urllib.request.Request(url, data = data, method = method))
    except:
        print("Error while showing image")
    return None

def value_range(dtype):
    if   dtype.name == "int8":    return (-128, 127)
    elif dtype.name == "uint8":   return (0, 255)
    elif dtype.name == "int16":   return (-32768, 32767)
    elif dtype.name == "uint16":  return (0, 65535)
    elif dtype.name == "int32":   return (-128, 127)
    elif dtype.name == "uint32":  return (0, 255)
    elif dtype.name == "float32": return (0, 1.0)
    elif dtype.name == "float64": return (0, 1.0)
    raise RuntimeError("Unsupported data type")


def imshow(name, image, vmin = None, vmax = None):
    """Shows an image. vmin and vmax are the minimum and maximum values for mapping or None to use default values."""
    vrange = value_range(image.dtype)
    if vmin is None:
        vmin = vrange[0]
    if vmax is None:
        vmax = vrange[1]

    data = Image.create_data(name, image, ImageValues(vmin = vmin, vmax = vmax))
    do_request(config.url_for("/image"), method = "POST", data = json.dumps(data).encode("utf-8"))


def close(name):
    do_request(config.url_for("/image/" + name), method = "DELETE")

def close_all():
    do_request(config.url_for("/images"), method = "DELETE")


def wait():
    wait_time = 0
    while True:
        resp = do_request(config.url_for("/wait/" + str(wait_time)), method = "GET")
        if resp is None:
            return
        msg = json.loads(resp.read())
        if msg["ready"] == True:
            return
        else:
            time = msg["time"]



def create_test_images():
    print("InitDB")
    imshow("test1", (255*np.random.rand(512, 512, 3)).astype(np.uint8))
    imshow("test2", np.random.rand(1024, 1024, 1))
    imshow("test3", np.random.rand(1024, 1024, 3))



from setuptools import setup

long_description = "An imshow showing the images in a browser."
	
setup(
    name = "wdisp",
    version = "1.0.0",
    description = "imshow for a browser",
    long_description = long_description, 

    url = "https://github.com/mgb4/wdisp",
    author = "Michael Burisch",
    author_email = "michael.burisch@gmx.com",
    
    keywords = "imshow browser",

    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3"
    ],
    
    packages = ["wdisp"],

    package_data = {
        "": ["wwwroot/*.*"]
    },


    python_requires = ">=3.4",
	
    install_requires = ["numpy", "PIL"],
    extras_require = {
        "Server": ["aiohttp"]
    }
)



import wdisp
import argparse
import time

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Image Server')
    parser.add_argument('--port', help = "Server port")
    args = parser.parse_args()
        
    wdisp.run_server_process(port = args.port)
    print("Server listening on", wdisp.config.root_url())
    print("Press Ctrl-C to stop program")
    try:
        while True:
            time.sleep(10)
    except KeyboardInterrupt:
        pass





import os.path


address = "http://localhost"
port = 4300

def wwwroot():
    return os.path.dirname(__file__) + "/wwwroot"

def root_url():
    return address + ":" + str(port) + "/"

def url_for(path):
    return address + ":" + str(port) + "/api" + path

